"""
Forward Chaining, K-Fold and Group K-Fold algorithms to split a given training dataset into train (X, y) and validation (Xcv, ycv) sets
"""

import numpy as np

def split_train_val_forwardChaining(sequence, numInputs, numOutputs, numJumps):
    """ Returns sets to train and cross-validate a model using forward chaining technique
    
    Parameters:
        sequence (array)  : Full training dataset
        numInputs (int)   : Number of inputs X and Xcv used at each training and validation
        numOutputs (int)  : Number of outputs y and ycv used at each training and validation
        numJumps (int)    : Number of sequence samples to be ignored between (X,y) sets

    Returns:
        X (2D array)    : Array of numInputs arrays used for training
        y (2D array)    : Array of numOutputs arrays used for training
        Xcv (2D array)  : Array of numInputs arrays used for cross-validation
        ycv (2D array)  : Array of numOutputs arrays used for cross-validation
        
    """
    
    X, y, Xcv, ycv = dict(), dict(), dict(), dict()
    j=2; # Tracks index of CV set at each train/val split
    
    # Iterate through all train/val splits
    while 1:
        start_ix=0; end_ix=0; startCv_ix=0; endCv_ix=0;
        X_it, y_it, Xcv_it, ycv_it = list(), list(), list(), list()
        i=0; # Index of individual training set at each train/val split
        
        # Iterate until index of individual training set is smaller than index of cv set
        while (i < j):
            ## TRAINING DATA
            start_ix = numJumps*i;
            end_ix = start_ix + numInputs;
            
            seq_x = sequence[start_ix:end_ix] 
            X_it.append(seq_x)
            seq_y = sequence[end_ix:end_ix+numOutputs]
            y_it.append(seq_y)
            
            i+=1;
          
        # Once val data crosses time series length return   
        if (((end_ix+numInputs)+numOutputs) > len(sequence)):
            break
        
        ## CROSS-VALIDATION DATA
        startCv_ix = end_ix;
        endCv_ix = end_ix + numInputs;
        
        seq_xcv = sequence[startCv_ix:endCv_ix] 
        Xcv_it.append(seq_xcv)
        seq_ycv = sequence[endCv_ix:endCv_ix+numOutputs]
        ycv_it.append(seq_ycv) 
            
        ## Add another train/val split 
        X[j-2] = np.array(X_it)
        y[j-2] = np.array(y_it)
        Xcv[j-2] = np.array(Xcv_it)
        ycv[j-2] = np.array(ycv_it)
        
        j+=1;
        
    if (len(X)==0 or len(Xcv)==0):
        print("The sequence provided does not has size enough to populate the return arrays")
   
    return X, y, Xcv, ycv


def split_train_val_kFold(sequence, numInputs, numOutputs, numJumps):
    """ Returns sets to train and cross-validate a model using K-Fold technique
    
    Parameters:
        sequence (array)  : Full training dataset
        numInputs (int)   : Number of inputs X and Xcv used at each training
        numOutputs (int)  : Number of outputs y and ycv used at each training
        numJumps (int)    : Number of sequence samples to be ignored between (X,y) sets

    Returns:
        X (2D array)    : Array of numInputs arrays used for training
        y (2D array)    : Array of numOutputs arrays used for training
        Xcv (2D array)  : Array of numInputs arrays used for cross-validation
        ycv (2D array)  : Array of numOutputs arrays used for cross-validation
        
    """
    
    X, y, Xcv, ycv = dict(), dict(), dict(), dict()
    j=2; # Tracks index of CV set at each train/val split
    theEnd = 0; # Flag to terminate function
    
    # Iterate until val set falls outside time series length
    while 1:
        start_ix=0; end_ix=0; startCv_ix=0; endCv_ix=0;
        X_it, y_it, Xcv_it, ycv_it = list(), list(), list(), list()
        i=0; # Index of individual training set at each train/val split
        n=0; # Number of numJumps
        
        # Iterate through all train/val splits
        while 1:
            if (i != j): 
                ## TRAINING DATA
                start_ix = endCv_ix + numJumps*n;
                end_ix = start_ix + numInputs;
                n +=1;

                # Leave train/val split loop once training data crosses time series length
                if end_ix+numOutputs > len(sequence):
                    break;

                seq_x = sequence[start_ix:end_ix] 
                X_it.append(seq_x)
                seq_y = sequence[end_ix:end_ix+numOutputs]
                y_it.append(seq_y)
            else:
                ## CROSS-VALIDATION DATA
                startCv_ix = end_ix;
                endCv_ix = end_ix + numInputs;
                n = 0;
                
                # Once val data crosses time series length exit tran/val split loop and return
                if endCv_ix+numOutputs > len(sequence):
                    theEnd = 1;
                    break;

                seq_xcv = sequence[startCv_ix:endCv_ix] 
                Xcv_it.append(seq_xcv)
                seq_ycv = sequence[endCv_ix:endCv_ix+numOutputs]
                ycv_it.append(seq_ycv)
            i+=1;
        
        # Only add a train/val split if the time series length has not been crossed
        if (theEnd == 1):
            break
        
        ## Add another train/val split 
        X[j-2] = np.array(X_it)
        y[j-2] = np.array(y_it)
        Xcv[j-2] = np.array(Xcv_it)
        ycv[j-2] = np.array(ycv_it)
        
        j+=1;
        
    if (len(X)==0 or len(Xcv)==0):
        print("The sequence provided does not has size enough to populate the return arrays")
            
    return X, y, Xcv, ycv


def split_train_val_groupKFold(sequence, numInputs, numOutputs, numJumps):
    """ Returns sets to train and cross-validate a model using group K-Fold technique
    
    Parameters:
        sequence (array)  : Full training dataset
        numInputs (int)   : Number of inputs X and Xcv used at each training
        numOutputs (int)  : Number of outputs y and ycv used at each training
        numJumps (int)    : Number of sequence samples to be ignored between (X,y) sets

    Returns:
        X (2D array)    : Array of numInputs arrays used for training
        y (2D array)    : Array of numOutputs arrays used for training
        Xcv (2D array)  : Array of numInputs arrays used for cross-validation
        ycv (2D array)  : Array of numOutputs arrays used for cross-validation
    
    """
    
    X, y, Xcv, ycv = dict(), dict(), dict(), dict()
    
    # Iterate through 5 train/val splits
    for j in np.arange(5):
        start_ix=0; end_ix=0; startCv_ix=0; endCv_ix=0;
        X_it, y_it, Xcv_it, ycv_it = list(), list(), list(), list()
        i=0; # Index of individual training set at each train/val split
        n=0; # Number of numJumps
        
        while 1: 
            if ((i+1+j)%(5) != 0):
                # TRAINING DATA
                start_ix = endCv_ix + numJumps*n;
                end_ix = start_ix + numInputs;
                n+=1;

                # Leave train/val split loop once training data crosses time series length
                if end_ix+numOutputs > len(sequence)-1:
                    break 

                seq_x = sequence[start_ix:end_ix] 
                X_it.append(seq_x)
                seq_y = sequence[end_ix:end_ix+numOutputs]
                y_it.append(seq_y)
            else:
                # CROSS-VALIDATION DATA
                startCv_ix = end_ix;
                endCv_ix = end_ix + numInputs;
                n=0;

                # Once val data crosses time series length return   
                if ((endCv_ix+numOutputs) > len(sequence)):
                    break

                seq_xcv = sequence[startCv_ix:endCv_ix] 
                Xcv_it.append(seq_xcv)
                seq_ycv = sequence[endCv_ix:endCv_ix+numOutputs]
                ycv_it.append(seq_ycv)  
                
            i+=1;
            
        ## Add another train/val split     
        X[j] = np.array(X_it)
        y[j] = np.array(y_it)
        Xcv[j] = np.array(Xcv_it)
        ycv[j] = np.array(ycv_it)
        
    if (len(X)==0 or len(Xcv)==0):
        print("The sequence provided does not has size enough to populate the return arrays")
            
    return X, y, Xcv, ycv



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['timeseries-cv']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.19.5,<1.20.0']

setup_kwargs = {
    'name': 'timeseries-cv',
    'version': '0.1.3',
    'description': 'Timeseries cross-validation for Neural Networks',
    'long_description': '# Time-Series Cross-Validation\n\nThis python package aims to implement Time-Series Cross Validation Techniques.\n\nThe idea is given a training dataset, the package will split it into Train, Validation and Test sets, by means of either Forward Chaining, K-Fold or Group K-Fold.\n\nAs parameters the user can not only select the number of inputs (n_steps_input) and outputs (n_steps_forecast), but also the number of samples (n_steps_jump) to jump in the data to train.\n\nThe best way to install the package is as follows:\n```\npip install git+https://github.com/DidierRLopes/TimeSeriesCrossValidation\n```\n-------------------------------------------------------------------------------\nThis can be seen more intuitively using the jupyter notebook: "example.ipynb"\nBelow you can find an example of the usage of each function for the following Time-Series:\n\ntimeSeries = array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26])\n\n-------------------------------------------------------------------------------\n## Split Train\n\n### split_train\n\nfrom TimeSeriesCrossValidation.splitTrain import split_train\n\nX, y = split_train(timeSeries, n_steps_input=4, n_steps_forecast=3, n_steps_jump=2)\n\n<img width="756" alt="train" src="https://user-images.githubusercontent.com/25267873/74095694-37600b80-4aec-11ea-979e-1bd50ed5851a.png">\n\n### split_train_variableInput\n\nfrom TimeSeriesCrossValidation.splitTrain import split_train_variableInput\n\nX, y = split_train_variableInput(timeSeries, minSamplesTrain=10, n_steps_forecast=3, n_steps_jump=3)\n\n![split_train_variableInput](https://user-images.githubusercontent.com/25267873/76267051-67243f80-6261-11ea-9eba-8a25fa810b06.png)\n\n-------------------------------------------------------------------------------\n## Split Train Val\n\n### split_train_val_forwardChaining\n\nfrom TimeSeriesCrossValidation.splitTrainVal import split_train_val_forwardChaining\n\nX, y, Xcv, ycv = split_train_val_forwardChaining(timeSeries, n_steps_input=4, n_steps_forecast=3, n_steps_jump=2)\n\n<img width="742" alt="trainVal - forwardChaining" src="https://user-images.githubusercontent.com/25267873/74094568-720d7800-4adb-11ea-8d69-7c1cbd6774c7.png">\n\n### split_train_val_kFold\n\nfrom TimeSeriesCrossValidation.splitTrainVal import split_train_val_kFold\n\nX, y, Xcv, ycv = split_train_val_kFold(timeSeries, n_steps_input=4, n_steps_forecast=3, n_steps_jump=2)\n\n<img width="743" alt="trainVal - kFold" src="https://user-images.githubusercontent.com/25267873/74094572-746fd200-4adb-11ea-91fd-93935d51982f.png">\n\n### split_train_val_groupKFold\nfrom TimeSeriesCrossValidation.splitTrainVal import split_train_val_groupKFold\n\nX, y, Xcv, ycv = split_train_val_groupKFold(timeSeries, n_steps_input=4, n_steps_forecast=3, n_steps_jump=2)\n\n<img width="744" alt="trainVal - groupKFold" src="https://user-images.githubusercontent.com/25267873/74094569-72a60e80-4adb-11ea-8345-1233b0a47e2e.png">\n\n-------------------------------------------------------------------------------\n## Split Train Val Test\n\n### split_train_val_test_forwardChaining\nfrom TimeSeriesCrossValidation.splitTrainValTest import split_train_val_test_forwardChaining\n\nX, y, Xcv, ycv, Xtest, ytest = split_train_val_test_forwardChaining(timeSeries, n_steps_input=4, n_steps_forecast=3, n_steps_jump=2)\n\n<img width="744" alt="trainValTest - forwardChaining" src="https://user-images.githubusercontent.com/25267873/74094566-6fab1e00-4adb-11ea-810d-e085518c3cb5.png">\n\n### split_train_val_test_kFold\n\nfrom TimeSeriesCrossValidation.splitTrainValTest import split_train_val_test_kFold\n\nX, y, Xcv, ycv, Xtest, ytest = split_train_val_test_kFold(timeSeries, n_steps_input=4, n_steps_forecast=3, n_steps_jump=2)\n\n<img width="745" alt="trainValTest - kFold" src="https://user-images.githubusercontent.com/25267873/74094570-73d73b80-4adb-11ea-94cd-5ab4d02c8cbf.png">\n\n### split_train_val_test_groupKFold\n\nfrom TimeSeriesCrossValidation.splitTrainValTest import split_train_val_test_groupKFold\n\nX, y, Xcv, ycv, Xtest, ytest = split_train_val_test_groupKFold(timeSeries, n_steps_input=4, n_steps_forecast=3, n_steps_jump=2)\n\n<img width="744" alt="trainValTest - groupKFold" src="https://user-images.githubusercontent.com/25267873/74094567-70dc4b00-4adb-11ea-994b-c3f1727f4b83.png">\n',
    'author': 'didier',
    'author_email': 'dro.lopes@campus.fct.unl.pt',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/DidierRLopes/timeseries-cv',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""
Algorithm to separate a given training set into input X and outputs y
"""

import numpy as np


def split_train(sequence, numInputs, numOutputs, numJumps):
    """ Returns sets to train a model
        i.e. X[0] = sequence[0], ..., sequence[numInputs]
             y[0] = sequence[numInputs+1], ..., sequence[numInputs+numOutputs]
             ...
             X[k] = sequence[k*numJumps], ..., sequence[k*numJumps+numInputs]
             y[k] = sequence[k*numJumps+numInputs+1], ..., sequence[k*numJumps+numInputs+numOutputs]
    
    Parameters:
        sequence (array)  : Full training dataset
        numInputs (int)   : Number of inputs X used at each training
        numOutputs (int)  : Number of outputs y used at each training
        numJumps (int)    : Number of sequence samples to be ignored between (X,y) sets

    Returns:
        X (2D array): Array of numInputs arrays.
                      len(X[k]) = numInputs
        y (2D array): Array of numOutputs arrays 
                      len(y[k]) = numOutputs
                      
    """
    
    X, y = list(), list()
    
    if (numInputs+numOutputs > len(sequence)):
        print("To have at least one X,y arrays, the sequence size needs to be bigger than numInputs+numOutputs")
        return X, y
    
    for i in range(len(sequence)):
        i = numJumps*i;
        end_ix = i + numInputs

        # Once train data crosses time series length return   
        if end_ix+numOutputs > len(sequence):
            break
        
        seq_x = sequence[i:end_ix] 
        X.append(seq_x)
        seq_y = sequence[end_ix:end_ix+numOutputs]
        y.append(seq_y)
        
    return X, y


def split_train_variableInput(sequence, minSamplesTrain, numOutputs, numJumps):
    """ Returns sets to train a model with variable input length
        i.e. X[0] = sequence[0], ..., sequence[minSamplesTrain]
             y[0] = sequence[0], ..., sequence[minSamplesTrain+numOutputs]
             ...
             X[k] = sequence[0], ..., sequence[k*numJumps+minSamplesTrain]
             y[k] = sequence[0], ..., sequence[k*numJumps+minSamplesTrain+numOutputs]
             
    Parameters:
        sequence (array)       : Full training dataset
        minSamplesTrain (int)  : Minimum number of inputs X used at each training
        numOutputs (int)       : Number of outputs y used at each training
        numJumps (int)         : Number of sequence samples to be jumped between (X,y) sets

    Returns:
        X (2D array): Array of input arrays.
                      len(X[k]) = minSamplesTrain + k*numJumps
        y (2D array): Array of numOutputs arrays 
                      len(y[k]) = minSamplesTrain+numOutputs+k*numJumps

    """
    
    end_ix=0; i=0;
    X, y = list(), list()
    
    if (minSamplesTrain+numOutputs > len(sequence)):
        print("To have at least one X,y arrays, the sequence size needs to be bigger than minSamplesTrain+numOutputs")
        return X, y
        
    # Iterate through all validation splits
    while 1:
        end_ix = minSamplesTrain + numJumps*i;
        
        # Training X
        seq_x = sequence[0:end_ix] 
        X.append(seq_x)
        
        # Training y
        seq_y = sequence[end_ix:end_ix+numOutputs]
        y.append(seq_y)

        i+=1;
          
        # Once val data crosses time series length return   
        if ((minSamplesTrain + numJumps*i + numOutputs) > len(sequence)):
            break
            
    return X, y
